/* retawq/resource.h - resource handling (network, cache, DNS, ...)
   This file is part of retawq (<http://retawq.sourceforge.net/>), a network
   client created by Arne Thomassen; retawq is basically released under certain
   versions of the GNU General Public License and WITHOUT ANY WARRANTY.
   Read the file COPYING for license details, README for program information.
   Copyright (C) 2001-2004 Arne Thomassen <arne@arne-thomassen.de>
*/

/* In the comments for data structure declarations in this file, "EXAC(thread)"
   meant that <thread> is the only thread which may access the marked member
   in any way ("exclusive access"). For a pointer, EXACD(thread) meant that
   only that thread may dereference the pointer. */

#ifndef __retawq_resource_h__
#define __retawq_resource_h__

#include "stuff.h"

#if HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif

#if OPTION_TLS == TLS_GNUTLS
#include <gnutls/gnutls.h>
typedef gnutls_session tTlsSession;
#elif OPTION_TLS == TLS_OPENSSL
#include <openssl/ssl.h>
typedef SSL* tTlsSession;
#elif OPTION_TLS == TLS_MATRIX
#include <matrixSsl.h>
typedef ssl_t* tTlsSession;
#elif OPTION_TLS == TLS_BUILTIN
struct __tTlsSession;
typedef struct __tTlsSession* tTlsSession;
#elif OPTION_TLS
#error Bad OPTION_TLS value
#endif

extern /*@observer@*/ const char strRedirection[];

extern const_after_init int fd_any2main_read, fd_any2main_write;
#if CONFIG_TG == TG_XCURSES
extern const_after_init int fd_xcurses2main_read, fd_xcurses2main_write;
#endif

extern int highest_select_fd;
#define is_fd_selectable(fd) ((fd) < FD_SETSIZE)


/* Resource protocols */

enum
{ rpUnknown = 0, rpDisabled = 1, rpHttp = 2, __rpFtp = 3, rpLocal = 4,
  rpAbout = 5, __rpFinger = 6, __rpLocalCgi = 7, __rpNntp = 8, __rpPop = 9,
  __rpPops = 10, rpCvs = 11, __rpGopher = 12, rpInfo = 13, __rpMailto = 14,
  __rpJavascript = 15, __rpHttps = 16, __rpFtps = 17, __rpExecextShell = 18
};
typedef unsigned char tResourceProtocol;
#define rpMax (18)
#define is_rp_nice(rp) ((rp) > rpDisabled)
#define MAXSCHEMESTRSIZE ((100) + 1)
  /* (just a value which is "big enough", even for _(strUnknown)) */

#if CONFIG_FINGER
#define rpFinger (__rpFinger)
#endif
#if CONFIG_GOPHER
#define rpGopher (__rpGopher)
#endif
#if CONFIG_FTP
#define rpFtp (__rpFtp)
#endif
#if OPTION_LOCAL_CGI
#define rpLocalCgi (__rpLocalCgi)
#endif
#if OPTION_NEWS
#define rpNntp (__rpNntp)
#endif
#if OPTION_POP
#define rpPop (__rpPop)
#if OPTION_TLS
#define rpPops (__rpPops)
#endif
#endif
#if OPTION_EDITOR
#define rpMailto (__rpMailto)
#endif
#if CONFIG_JAVASCRIPT
#define rpJavascript (__rpJavascript)
#endif

#if OPTION_TLS
#define rpHttps (__rpHttps)
#if CONFIG_FTP
#define rpFtps  (__rpFtps)
#endif
#endif

#if OPTION_EXECEXT & EXECEXT_SHELL
#define rpExecextShell (__rpExecextShell)
#endif

my_enum1 enum
{ rpfNone = 0, rpfIsLocallike = 0x01, rpfIsHttplike = 0x02,
  rpfIsFtplike = 0x04, rpfIsPoplike = 0x08, rpfIsTlslike = 0x10,
  rpfUsesAuthority = 0x20, rpfIsPathHierarchical = 0x40, rpfUiToupper = 0x80,
  rpfUiAlternate = 0x100
} my_enum2(unsigned short) tRpFlags;

typedef struct
{ const char* scheme;
  tResourceProtocol final_rp;
  tRpFlags flags;
} tRpData;

extern const tRpData rp_data[rpMax + 1];

#define __has_rpflag(rp, rpf) (rp_data[(rp)].flags & (rpf))
#define is_locallike(rp) __has_rpflag((rp), rpfIsLocallike)
#define is_httplike(rp) __has_rpflag((rp), rpfIsHttplike)
#if CONFIG_FTP
#define is_ftplike(rp) __has_rpflag((rp), rpfIsFtplike)
#endif
#if OPTION_POP
#define is_poplike(rp) __has_rpflag((rp), rpfIsPoplike)
#endif
#if OPTION_TLS
#define is_tlslike(rp) __has_rpflag((rp), rpfIsTlslike)
#endif


/* Resource data */

enum { rckUnknown = 0, rckPlainText = 1, rckHtml = 2 };
typedef unsigned char tResourceContentKind;

my_enum1 enum
{ rfNone = 0, rfFinal = 0x01, rfNeedUnmap = 0x02, rfHttp11 = 0x04,
  rfHttpChunked = 0x08, rfCannotReload = 0x10, rfPost = 0x20,
  rfIsRedirection = 0x40, rfHtmlRedirection = 0x80, rfIsEmbedded = 0x100,
  rfIsEnforced = 0x200, rfActivityWatched = 0x400,
  rfActivityUnwatchable = 0x800, rfDownload = 0x1000
#if CONFIG_FTP
  , rfFtpLastLine = 0x2000, rfFtpTryingEpsv = 0x4000, rfFtpDir = 0x8000,
  rfFtpSentAccount = 0x10000, rfFtpShakyNext = 0x20000
#if OPTION_TLS
  , rfFtpsDataClear = 0x40000, rfFtpsShakyProt = 0x80000
#define rfFtpShakyAny (rfFtpShakyNext | rfFtpsShakyProt)
#else
#define rfFtpShakyAny (rfFtpShakyNext)
#endif
#endif
#if OPTION_NEWS
  , rfNntpHeaderSeries = 0x100000
#endif
#if CONFIG_ANY_MAIL
  , rfMailbox = 0x200000
#endif
#if OPTION_COOKIES
  , rfCookieSender = 0x400000, rfCookieStorer = 0x800000 /* RFC2965, 6.1, *2 */
#define rfCookieAnything (rfCookieSender | rfCookieStorer)
#endif
} my_enum2(unsigned int) tResourceFlags;
/* (rfNeedUnmap implies that the content data is stored indirectly) */

enum
{ rsUnknown = 0, rsError = 1, rsConnecting = 2, rsReading = 3, rsComplete = 4,
  rsStopped = 5
};
typedef signed char tResourceState;
#define MAX_RESOURCE_STATE (5)
extern /*@observer@*/ const char* const
  strResourceState[MAX_RESOURCE_STATE + 1];

my_enum1 enum
{ rchUnknown = 0, rchConnected = 1
#if CONFIG_FTP
  , rchFtpUser = 2, rchFtpPassword = 3, rchFtpAccount = 4, rchFtpPasv = 5,
  rchFtpRequest = 6
#if OPTION_TLS
  , rchFtpTlsAuthTls = 7, rchFtpTlsAuthSsl = 8, rchFtpTlsPbsz = 9,
  rchFtpTlsProt = 10
#endif
#endif
#if OPTION_NEWS
  , rchNntpModeReader = 11, rchNntpGetGroups = 12, rchNntpFetchArticle = 13,
  rchNntpFetchHeader = 14, rchNntpSelectGroup = 15
#endif
#if OPTION_POP
  , rchPopUser = 16, rchPopPass = 17, rchPopApop = 18, rchPopStat = 19,
  rchPopTop = 20, rchPopUidl = 21, rchPopRetr = 22
#if OPTION_TLS
  , rchPopStls = 23
#endif
#endif
} my_enum2(unsigned char) tResourceCommandHandshake;

enum
{ cpsUnknown = 0, cpsData = 1, cpsSize = 2, cpsCrlf1 = 3, cpsSkipSizeline = 4,
  cpsCr2 = 5, cpsCrlf2 = 6, cpsSkipTrailerA = 7, cpsSkipTrailerB = 8,
  cpsSkipTrailerC = 9, cpsSkipTrailerD = 10
};
typedef unsigned char tChunkParsingState; /* for HTTP/1.1 chunk encoding */

enum
{ reFine = 0, reKind = 1, reSocket = 2, reConnect = 3, reRefused = 4,
  reNetwork = 5, reDns = 6, reFile = 7, reUri = 8, reServerClosed = 9,
  reHandshake = 10, rePortnumber = 11, reProtocol = 12, reTimeout = 13,
  reResponse = 14, reLogin = 15, reRedirection = 16, reRandom = 17, reTls = 18,
  reProxyAuth = 19, rePipe = 20, reFork = 21, reConfigForbids = 22,
  reExec = 23, reProtDisabled = 24, reHostname = 25, reTmofd = 26
};
typedef unsigned char tResourceError;
#define MAX_RESOURCE_ERROR (26)
extern /*@observer@*/ const char* const
  strResourceError[MAX_RESOURCE_ERROR + 1];

#if OPTION_TLS
enum
{ teFine = 0, teUnknown = 1, teSessionExpired = 2, teFatalAlert = 3,
  teFile = 4, teData = 5, teOom = 6, teCipher = 7, teCompression = 8,
  teCrypt = 9, teSrp = 10, tePk = 11, tePacket = 12, teUnimplemented = 13,
  teHandshake = 14, teInit = 15
};
typedef unsigned char tTlsError;
#define MAX_TLS_ERROR (15)
extern /*@observer@*/ const char* const strTlsError[MAX_TLS_ERROR + 1];
#define is_tls_error_expressive(te) ((te) > teUnknown) /* for UI */
#endif

typedef struct tSaveAs
{ struct tSaveAs* next;
  int fd;
} tSaveAs;

typedef unsigned short tServerStatusCode; /* 0 or roughly 100..699 */
typedef signed char tTlHeaderState; /*transport-level header state (misnomer)*/

#if CONFIG_DEBUG
typedef unsigned int tResourceVersion;
#endif

#define UNKNOWN_CONTENTLENGTH ((size_t) (~((size_t) 0)))

struct tConnection;
struct tResourceRequest;
typedef signed int tReqlistNum; /* ("signed" for simplicity only) */
struct tResourceCacheEntry;

#if 1 || CONFIG_FTP || OPTION_NEWS || OPTION_POP || OPTION_TRAP
/* (the "1 ||" is for HTTP nowadays) */
#define USE_RPSD 1
struct tRpsdGeneric; /* resource-protocol-specific data */
#else
#define USE_RPSD 0
#endif

typedef struct
{ struct tResourceRequest** reqlist;
  struct tConnection *cconn, *dconn; /* control/data; EXAC(resource) */
  char *uri, *path, *query, *post, *username, *password;
  char* content; /* first content block */
  char* lastcontent; /* last received content block; EXAC(resource) */
  void* tree; /* built from <content> (e.g. tree of HTML tags) */
  const tConfigProxy* proxy;
  tActiveElementBase* aebase;
  tHtmlForm* form;
  char* lhpp_content; /* last HTML parser position: content block */
  char* redirection; /* for HTTP 3xy status codes */
  struct tHostPortProtInfo *actual_hppi, *textual_hppi; /* EXAC(resource) */
    /* (might differ, e.g. when proxies are used; textual_... always holds
        stuff from the URI) */
  struct tResourceCacheEntry* cache_entry; /* EXAC(resource) */
  tSaveAs* save_as; /* for saving to disk; EXAC(resource) */
#if USE_RPSD
  struct tRpsdGeneric* rpsd; /* EXAC(resource) */
#endif
  tObjectIdentifier id;
  size_t nominal_contentlength;
  size_t bytecount; /* roughly "number of bytes in document" */
  size_t chunksize_left; /* for HTTP 1.1, chunked encoding; EXAC(resource) */
  size_t lhpp_byte; /* last HTML parser position: byte within content block */
  tResourceFlags flags;
  tReqlistNum reqlist_num, reqlist_maxnum;
  tActiveElementNumber aenum, aemax;
  tHtmlFormNumber hfnum, hfmax;
  int download_fd;
#if CONFIG_DEBUG
  int err; /* OS error code */
  tResourceVersion version; /* to be increased whenever the content changes */
#endif
  tPortnumber textual_portnumber; /* (in network byte order) */
  tServerStatusCode server_status_code; /* (FTP terminology: reply code) */
  tTlHeaderState tlheaderstate; /* EXAC(resource) */
  tResourceProtocol protocol;
  tResourceState state;
  tResourceCommandHandshake handshake;
  tResourceContentKind kind;
  tChunkParsingState cps; /* EXAC(resource) */
  tResourceError error;
#if OPTION_TLS
  tTlsError tls_error; /* used if ->error == reTls */
#endif
} tResource;


/* Content blocks */

#define CONTENTBLOCK_NEXT(base) ( *((char**)((char*)(base + 0))) )

#define __CONTENTBLOCK_INDIRECTIONBIT ((size_t) (1 << 30))
  /* (We abuse this bit in the "usable size" field to indicate whether the
      content data is addressed directly or indirectly.) */
#define __CONTENTBLOCK_USABLESIZE(base) \
  ( *((size_t*)((char*)(base + sizeof(char**)))) )
#define CONTENTBLOCK_USABLESIZE(base) \
  (__CONTENTBLOCK_USABLESIZE(base) & (~__CONTENTBLOCK_INDIRECTIONBIT))
#define CONTENTBLOCK_DATA_IS_INDIRECT(base) \
  ( (__CONTENTBLOCK_USABLESIZE(base) & __CONTENTBLOCK_INDIRECTIONBIT) != 0 )

#define CONTENTBLOCK_USEDSIZE(base) \
  ( *((size_t*)((char*)(base + sizeof(char**) + sizeof(size_t*)))) )
#define __CONTENTBLOCK_DATA(base) \
  ((char*) (base + sizeof(char**) + 2 * sizeof(size_t*)))
#define __CONTENTBLOCK_DATAPTR(base) \
  (*((char**) __CONTENTBLOCK_DATA(base)))
#define CONTENTBLOCK_DATA(base) \
  ( (CONTENTBLOCK_DATA_IS_INDIRECT(base)) \
    ? (__CONTENTBLOCK_DATAPTR(base)) : (__CONTENTBLOCK_DATA(base)) )


/* Resource requests */

my_enum1 enum
{ rraCancel = 0, rraLoad = 1, rraReload = 2, rraEnforcedReload = 3,
  rraInteractiveConn = 4
} my_enum2(unsigned char) tResourceRequestAction;

#define is_loadrra(rra) \
  ( ((rra) == rraLoad) || ((rra) == rraReload) || ((rra) == rraEnforcedReload))

my_enum1 enum
{ rrsUnknown = 0, rrsError = 1, rrsPreparedByMain = 2, rrsAttachedResource = 3,
  rrsStopped = 4
#if CONFIG_ASYNC_DNS
  , rrsDnsLookup = 5
#endif
} my_enum2(unsigned char) tResourceRequestState;
#define MAX_RESOURCE_REQUEST_STATE (5)
extern /*@observer@*/ const char* const
  strResourceRequestState[MAX_RESOURCE_REQUEST_STATE + 1];

my_enum1 enum
{ rrfNone = 0, rrfResourceChanged = 0x01, rrfMayOverwrite = 0x02,
  rrfPost = 0x04, rrfTryIpv6 = 0x08, rrfIsRedirection = 0x10,
  rrfIsEmbedded = 0x20, rrfDownload = 0x40
#if CONFIG_FTP
  , rrfFtpTypeA = 0x80, rrfFtpTypeD = 0x100, rrfFtpTypeI = 0x200
#endif
} my_enum2(unsigned short) tResourceRequestFlags;

struct tWindowView;
struct tCachedHostInformation;

typedef struct tResourceRequest
{ struct tResourceRequest *next /* , *child */ ;
    /* for lists and trees, e.g. framesets; EXAC(main) */
  char *uri, *hostname, *username, *password, *path, *query, *post;
  const tConfigProxy* proxy; /* EXAC(resource) */
  tResource* resource; /* the result of the request */
  struct tWindowView* window_view; /* EXAC(main) */
  struct tCachedHostInformation* lookup; /* EXAC(resource) */
  tObjectIdentifier id;
  int download_fd;
  tPortnumber textual_portnumber; /* (in network byte order) */
  tResourceRequestFlags flags;
  tResourceRequestAction action;
  tResourceRequestState state;
  tResourceProtocol protocol;
  tResourceError error;
} tResourceRequest;


/* Communication between main.c and resource.c */

enum
{ /* from main.c to resource.c: */
  mckCalledSelect = 0, mckStart = 1, mckFds = 2, mckRemovingRequest = 3,
  mckStartSaveAs = 4, mckStopActivity = 5,
#if CONFIG_USER_QUERY
  mckUserResponse = 6,
#endif
#if OPTION_EXECEXT & EXECEXT_SHELL
  mckExecextShell = 7,
#endif
  /* from resource.c to main.c: */
  mckResource = 8, mckRequest = 9
#if CONFIG_USER_QUERY
  , mckUserQuery = 10
#endif
};
typedef unsigned char tMainCallbackKind;
typedef struct { int selectcount; } tMainCallbackFds; /* for mckFds */
extern fd_set fds_read, fds_write, fds_r, fds_w, fds_e;
extern void main_callback(void*, tMainCallbackKind); /* in main.c */
extern void resource_callback(void*, tMainCallbackKind); /* in resource.c */

#if CONFIG_USER_QUERY

my_enum1 enum
{ mifNone = 0, mifQueryFailed = 0x01, mifUserCancelled = 0x02,
  mifObjectVanished = 0x04, mifProxyrelated = 0x08,
  mifPriorLoginAttemptFailed = 0x10, mifUsername = 0x20, mifPassword = 0x40,
  mifAccount = 0x80, mifSharedSecret = 0x100
#if CONFIG_FTP && OPTION_TLS
  , mifFtpsDataclear = 0x200
#endif
} my_enum2(unsigned short) tMissingInformationFlags;
#define mifSet (mifUsername | mifPassword | mifAccount | mifSharedSecret)

struct tUserQuery;
typedef void (*tUserQueryCallback)(struct tUserQuery*);
typedef struct tUserQuery
{ tUserQueryCallback callback; /* usually a function in resource.c */
  tResource* resource;
  struct tHostPortProtInfo* hppi; /* EXAC(resource) */
  const char* hostname; /* just a pointer copy, no strdup()ing necessary */
  tObjectIdentifier id; /* from tResource */
  tPortnumber portnumber; /* (in network byte order) */
  tMissingInformationFlags mif;
} tUserQuery;

#else

typedef struct { unsigned char dummy; } tUserQuery;

#endif

#if OPTION_EXECEXT & EXECEXT_SHELL

my_enum1 enum
{ esfNone = 0, esfReadStdout = 0x01, esfReadStderr = 0x02,
  esfEnforceHtml = 0x04
} my_enum2(unsigned char) tExecextShellFlags;

typedef struct
{ tResourceRequest* request;
  const char *command, *writedata;
  size_t writedata_size;
  tExecextShellFlags esf;
} tExecextShellData;

#endif


/* Functions */

extern tPortnumber rp2portnumber(tResourceProtocol);
extern void rp2scheme(tResourceProtocol, /*@out@*/ char*);
#if CONFIG_BLOAT & BLOAT_SSC
extern /*@observer@*/ const char* ssc2info(tResourceProtocol,
  tServerStatusCode);
#else
#define ssc2info(rp, ssc) (strEmpty)
#endif

#define USE_S2U ((CONFIG_BLOAT & BLOAT_UIIP) || (CONFIG_DEBUG))
#if USE_S2U
extern tBoolean resource_ui_conn_ip(const tResource*, /*@out@*/ char*, size_t);
#else
#define resource_ui_conn_ip(res, buf, size) (falsE)
#endif
extern void does_not_return fatal_tmofd(int);
extern void my_FD_SET(int, fd_set*);
extern tBoolean my_try_FD_SET(int, fd_set*);

#if OPTION_TLS
extern tBoolean resource_is_tls_handshaking(const tResource*);
#endif

extern void resource_initialize(void);

#endif /* #ifndef __retawq_resource_h__ */
